/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "gwf_source.h"
#include "gwf_tools.h"
#include "gwf_project.h"
#include "gwf_license.h"
#include "gwf_error.h"
#include <gtk/gtk.h>
#include <string.h>

static gint		prior_compare			(gconstpointer a,gconstpointer b);
static gint             name_compare			(gconstpointer a,gconstpointer b);
static gint		check_name_compare		(gconstpointer a,gconstpointer b);

static void		free_write_func 		(GwfWriteData *data);
static void		free_check_func			(GwfCheckData *data);
static void		free_write_func_foreach 	(gpointer data,gpointer user_data);
static void		exec_write_func_foreach 	(gpointer data,gpointer user_data);
static void		free_check_func_foreach		(gpointer data,gpointer user_data);
static void		check_source_foreach		(gpointer data,gpointer user_data);
static void		enable_functions		(GwfSource *srcwrt);
static void             enable_func_foreach		(GwfProjData *data,gpointer user_data);

static void		write_generator_title		(GString *src);
static void		write_copyright			(GString *src,gchar* author,gchar *email);
static void		write_license			(GString *src,GwfLicList *liclist,GwfProject *project);
static void		write_hdef_start		(GString *src,gchar* hfile);
static void		write_hdef_end			(GString *src,gchar* hfile);
static void		write_includes			(GString *src,gchar *hfile);


GwfSource*
gwf_source_write_new (void)
{
  GwfSource	*srcwrt;

  srcwrt = g_malloc(sizeof(GwfSource));
  srcwrt->checks = NULL;
  srcwrt->c_funcs = NULL;
  srcwrt->h_funcs = NULL;
  srcwrt->src = g_string_new(NULL);
  srcwrt->project = NULL;

  return srcwrt;
}

void
gwf_source_write_free (GwfSource *srcwrt)
{
  g_return_if_fail(srcwrt != NULL);

  g_list_foreach(srcwrt->c_funcs,free_write_func_foreach,NULL);
  g_list_foreach(srcwrt->h_funcs,free_write_func_foreach,NULL);
  g_list_foreach(srcwrt->checks,free_check_func_foreach,NULL);
  g_list_free(srcwrt->c_funcs);
  g_list_free(srcwrt->h_funcs);
  g_list_free(srcwrt->checks);
  g_string_free(srcwrt->src,TRUE);
  g_free(srcwrt);
}

void
gwf_source_schedule_func (GwfSource *srcwrt,gchar *name,GwfWritePrior prior,GwfSrcWrtFunc func,GwfWriteFile file)
{
  GwfWriteData	*data;

  g_return_if_fail(srcwrt != NULL);
  g_return_if_fail(name != NULL);
  g_return_if_fail(func != NULL);

  data = g_malloc(sizeof(GwfWriteData));
  data->prior = prior;
  data->func = func;
  data->name = g_strdup(name);
  data->enabled = TRUE;

  switch (file)
  {
    case HFILE	:	srcwrt->h_funcs = g_list_insert_sorted(srcwrt->h_funcs,(gpointer)data,prior_compare);
    			break;
    case CFILE	:	srcwrt->c_funcs = g_list_insert_sorted(srcwrt->c_funcs,(gpointer)data,prior_compare);
    			break;
    default	:	g_warning("Unknown file type!!!");
    		        return;
  }
}

void
gwf_source_unschedule_by_name (GwfSource *srcwrt,gchar *name)
{
  GList		*listelem;

  g_return_if_fail(srcwrt != NULL);
  g_return_if_fail(name != NULL);


  while ((listelem = g_list_find_custom(srcwrt->h_funcs,(gpointer)name,name_compare)) != NULL)
  {
    free_write_func((GwfWriteData*)listelem->data);
    g_list_remove_link(srcwrt->h_funcs,listelem);
  }

  while ((listelem = g_list_find_custom(srcwrt->c_funcs,(gpointer)name,name_compare)) != NULL)
  {
    free_write_func((GwfWriteData*)listelem->data);
    g_list_remove_link(srcwrt->c_funcs,listelem);
  }
}

gint
gwf_source_add_check (GwfSource *srcwrt,gchar *name,GwfSrcCheckFunc func)
{
  GwfCheckData	*tmp;

  g_return_val_if_fail(srcwrt != NULL,0);
  g_return_val_if_fail(name != NULL,0);
  g_return_val_if_fail(func != NULL,0);

  if (g_list_find_custom(srcwrt->checks,(gpointer)name,check_name_compare))
    return 0;

  tmp = g_malloc(sizeof(GwfCheckData));
  tmp->name = g_strdup(name);
  tmp->enabled = TRUE;
  tmp->func = func;

  srcwrt->checks = g_list_append(srcwrt->checks,(gpointer)tmp);

  return 1;
}

void
gwf_source_del_check (GwfSource *srcwrt,gchar *name)
{
  GList		*link;

  g_return_if_fail(srcwrt != NULL);
  g_return_if_fail(name != NULL);

  link = g_list_find_custom(srcwrt->checks,(gpointer)name,check_name_compare);
  if (!link)
    return;

  srcwrt->checks = g_list_remove_link(srcwrt->checks,link);
  free_check_func((GwfCheckData*)link->data);
  g_free(link);
}

void
gwf_source_set_by_name (GwfSource *srcwrt,gchar *name,gboolean enabled)
{
  GList		*listelem;

  g_return_if_fail(srcwrt != NULL);
  g_return_if_fail(name != NULL);

  for (listelem = srcwrt->h_funcs; listelem ; listelem = listelem ->next)
  {
    if (strcmp(((GwfWriteData*)listelem->data)->name,name) == 0)
      ((GwfWriteData*)listelem->data)->enabled = enabled;
  }

  for (listelem = srcwrt->c_funcs; listelem ; listelem = listelem ->next)
  {
    if (strcmp(((GwfWriteData*)listelem->data)->name,name) == 0)
      ((GwfWriteData*)listelem->data)->enabled = enabled;
  }

  for (listelem = srcwrt->checks; listelem ; listelem = listelem ->next)
  {
    if (strcmp(((GwfCheckData*)listelem->data)->name,name) == 0)
      ((GwfCheckData*)listelem->data)->enabled = enabled;
  }
}

gint
gwf_source_write_check(GwfSource *srcwrt,GwfProject *project,GwfError *error)
{
  g_return_val_if_fail(srcwrt != NULL,0);
  g_return_val_if_fail(project != NULL,0);

  srcwrt->result = 1;
  srcwrt->project = project;
  srcwrt->error = error;

  enable_functions(srcwrt);

  g_list_foreach(srcwrt->checks,check_source_foreach,(gpointer)srcwrt);

  return srcwrt->result;
}

gchar*
gwf_source_write_h (GwfSource *srcwrt,GwfProject *project,GwfLicList *liclist)
{
  g_return_val_if_fail(srcwrt != NULL,NULL);
  g_return_val_if_fail(project != NULL,NULL);

  g_string_free(srcwrt->src,TRUE);

  srcwrt->src = g_string_new(NULL);
  srcwrt->project = project;

  enable_functions(srcwrt);

  write_generator_title(srcwrt->src);
  write_copyright(srcwrt->src,gwf_project_get_author(project),gwf_project_get_email(project));
  write_license(srcwrt->src,liclist,project);
  write_hdef_start(srcwrt->src,gwf_project_get_hfile(project));

  g_list_foreach(srcwrt->h_funcs,exec_write_func_foreach,(gpointer)srcwrt);

  write_hdef_end(srcwrt->src,gwf_project_get_hfile(project));

  return g_strdup(srcwrt->src->str);
}

gchar*
gwf_source_write_c (GwfSource *srcwrt,GwfProject *project,GwfLicList *liclist)
{
  g_return_val_if_fail(srcwrt != NULL,NULL);
  g_return_val_if_fail(project != NULL,NULL);

  g_string_free(srcwrt->src,TRUE);

  srcwrt->src = g_string_new(NULL);
  srcwrt->project = project;

  enable_functions(srcwrt);

  write_generator_title(srcwrt->src);
  write_copyright(srcwrt->src,gwf_project_get_author(project),gwf_project_get_email(project));
  write_license(srcwrt->src,liclist,project);
  write_includes(srcwrt->src,gwf_project_get_hfile(project));

  g_list_foreach(srcwrt->c_funcs,exec_write_func_foreach,(gpointer)srcwrt);

  return g_strdup(srcwrt->src->str);
}

/***************************/
/* Function list functions */
/***************************/

static gint
prior_compare(gconstpointer a,gconstpointer b)
{
  const GwfWriteData	*elem1 = a;
  const GwfWriteData	*elem2 = b;

  return (elem1->prior+1 < elem2->prior) ? -1 : (elem1->prior+1 == elem2->prior) ? 0 : 1;
}

static gint
name_compare(gconstpointer a,gconstpointer b)
{
  return strcmp(((GwfWriteData*)a)->name,(gchar*)b);
}

static gint
check_name_compare(gconstpointer a,gconstpointer b)
{
  return strcmp(((GwfCheckData*)a)->name,(gchar*)b);
}

static void
check_source_foreach(gpointer data,gpointer user_data)
{
  GwfCheckData		*chkdat = data;
  GwfSource		*srcwrt = user_data;
  gint			tmp;

  if (!chkdat->enabled)
    return;

  tmp = chkdat->func(srcwrt->project,srcwrt->error);

  if (srcwrt->result)
    srcwrt->result = tmp;
}

/**************************/
/* GwfWriteData functions */
/**************************/
static void
free_write_func (GwfWriteData *data)
{
  g_free(data->name);
  g_free(data);
}

static void
free_check_func (GwfCheckData *data)
{
  g_free(data->name);
  g_free(data);
}

static void
free_write_func_foreach (gpointer data,gpointer user_data)
{
  free_write_func((GwfWriteData*)data);
}

static void
free_check_func_foreach (gpointer data,gpointer user_data)
{
  free_check_func((GwfCheckData*)data);
}

static void
exec_write_func_foreach (gpointer data,gpointer user_data)
{
  GwfWriteData		*wrtdat = data;
  GwfSource		*srcwrt = user_data;

  if (wrtdat->enabled)
    wrtdat->func(srcwrt->project,srcwrt->src);
}

static void
enable_functions(GwfSource *srcwrt)
{
  GwfDataList	*datalist;

  g_return_if_fail(srcwrt != NULL);
  g_return_if_fail(srcwrt->project != NULL);

  datalist = gwf_project_get_data_list(srcwrt->project);
  g_return_if_fail(datalist != NULL);

  gwf_data_list_foreach(datalist,enable_func_foreach,(gpointer)srcwrt);
}

static void
enable_func_foreach(GwfProjData *data,gpointer user_data)
{
  GwfSource	*srcwrt = user_data;

  gwf_source_set_by_name(srcwrt,data->name,gwf_proj_data_get_active(data));
}

/**********************************/
/* General source write functions */
/**********************************/

static void
write_generator_title (GString *src)
{
  g_return_if_fail(src != NULL);
  g_string_append(src,"/* Gtk Widget Factory generated widget\n");
  g_string_append(src," * http://gwf.sourceforge.net\n");
  g_string_append(src," * Copyright (c) 1999-2000 Jeroen Benckhuijsen\n");
  g_string_append(src,"*/\n\n");
}

static void
write_copyright(GString *src,gchar *author,gchar *email)
{
  g_return_if_fail(src != NULL);

  g_string_sprintfa(src,"/* This file is Copyrighted (C) 1999-2000 by %s",(author) ? author : "UNKNOWN AUTHOR");
  if (email)
    g_string_sprintfa(src," (%s)",email);
  g_string_append(src,"\n * Please read the license agreement below\n");
  g_string_append(src,"*/\n\n");
}

static void
write_license(GString *src,GwfLicList *liclist,GwfProject *project)
{
  GwfLicense	*license;
  gchar		*name;

  g_return_if_fail(src != NULL);
  g_return_if_fail(liclist != NULL);
  g_return_if_fail(project != NULL);

  name = gwf_project_get_license(project);
  if (!name)
    name = liclist->std;

  license = gwf_liclist_get(liclist,name);
  if (!license)
    return;

  g_string_append(src,license->text);
}

static void
write_hdef_start(GString *src,gchar *hfile)
{
  gchar		*hfileup;
  gchar		*dot;

  g_return_if_fail(src != NULL);
  g_return_if_fail(hfile != NULL);

  hfileup = g_strdup(hfile);
  g_strup(hfileup);
  dot = strchr(hfileup,'.');
  *dot = '_';

  g_string_sprintfa(src,"#ifndef __%s__\n",hfileup);
  g_string_sprintfa(src,"#define __%s__\n",hfileup);
  g_string_append(src,"\n");
  g_string_append(src,"#ifdef __cplusplus\n");
  g_string_append(src,"  extern \"C\" {\n");
  g_string_append(src,"#endif\n\n");
  g_string_append(src,"#include <gtk/gtk.h>\n\n");

  g_free(hfileup);
}

static void
write_hdef_end(GString *src,gchar *hfile)
{
  gchar		*hfileup;
  gchar		*dot;

  g_return_if_fail(src != NULL);
  g_return_if_fail(hfile != NULL);

  hfileup = g_strdup(hfile);
  g_strup(hfileup);
  dot = strchr(hfileup,'.');
  *dot = '_';

  g_string_append(src,"#ifdef __cplusplus\n");
  g_string_append(src,"  }\n");
  g_string_append(src,"#endif\n");
  g_string_append(src,"\n");
  g_string_sprintfa(src,"#endif /* __%s__ */\n",hfileup);

  g_free(hfileup);
}

static void
write_includes(GString *src,gchar *hfile)
{
  g_return_if_fail(src != NULL);
  g_return_if_fail(hfile != NULL);

  g_string_append(src,"#ifdef HAVE_CONFIG_H\n");
  g_string_append(src,"  #include <config.h>\n");
  g_string_append(src,"#endif\n");
  g_string_append(src,"\n");
  g_string_append(src,"#include <gtk/gtk.h>\n");
  g_string_sprintfa(src,"#include \"%s\"\n",hfile);
}
